#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Collect unstructured content (text) from specific web page sources.

see copyright/license https://github.com/DerwenAI/strwythura/README.md
"""

import ssl
import unicodedata

from bs4 import BeautifulSoup
import requests
import requests_cache


class Scraper:
    """
A simple HTML scraper which caches content.
    """
    SCRAPE_HEADERS: dict[ str, str ] = {
        "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/39.0.2171.95 Safari/537.36",  # pylint: disable=C0301
    }

    CHAR_MAP: dict[ int, str ] = {
        226: "'",
        8220: '"',
        8221: '"',
        8216: "'",
        8217: "'",
    }

    SKIP_CHARS: set[ int ] = set([
        128,
        148,
        153,
        156,
        157,
        8203,
    ])


    def __init__ (
        self,
        config: dict,
        ) -> None:
        """
Constructor.
        """
        self.config: dict = config
        self.session: requests_cache.CachedSession = self.get_cache()


    def get_cache (
        self,
        ) -> requests_cache.CachedSession:
        """
Build a URL request cache session, optionally loading any
previous serialized cache from disk.
        """
        session: requests_cache.CachedSession = requests_cache.CachedSession(
            backend = requests_cache.SQLiteCache(
                self.config["scraper"]["cache_path"],
            ),
        )

        session.settings.expire_after = self.config["scraper"]["cache_expire"]

        return session


    def scrub_text (
        self,
        text: str | None,
        ) -> str | None:
        """
Scrub text of non-printable characters, typesetting artifacts, UTF-8 errors, etc.
Courtesy of <https://github.com/DerwenAI/pytextrank>
        """
        if text is None:
            return None

        # explode the string into characters to be mapped
        exploded = []

        for char in text:
            rach: int = ord(char)
            swap: str | None = self.CHAR_MAP.get(rach)

            if swap is not None:
                exploded.append(swap)
            elif rach not in self.SKIP_CHARS:
                exploded.append(char)

        min_scrub: str = unicodedata.normalize(
            "NFKD",
            "".join(exploded),
        ).strip()

        #max_scrub: str = min_scrub.encode("ascii", "ignore").decode("utf-8").strip()

        return min_scrub


    def scrape_html (
        self,
        url: str,
        ) -> list[ str ]:
        """
A simple web page content scraper, which returns a list of text paragraphs.
        """
        response: requests.Response = self.session.get(
            url,
            verify = ssl.CERT_NONE,
            timeout = 10,
            allow_redirects = True,
            headers = self.SCRAPE_HEADERS,
        )

        soup: BeautifulSoup = BeautifulSoup(
            response.text,
            features = "lxml",
        )

        return [
            self.scrub_text(para.text.strip())  # type: ignore
            for para in soup.find_all("p")
        ]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
A wrapper for the `pyInstrument` statistical call stack profiler.

see copyright/license https://github.com/DerwenAI/strwythura/README.md
"""

import os

from psutil._common import bytes2human
import psutil
import pyinstrument


class Profiler:  # pylint: disable=R0903
    """
Use statistical call stack sampling to augment LLM observability.
    """

    def __init__(
        self,
        ) -> None:
        """
Constructor.
        """
        self.profiler: pyinstrument.Profiler = pyinstrument.Profiler()
        self.profiler.start()


    def analyze (
        self,
        ) -> None:
        """
Analyze and report the performance measures.
        """
        self.profiler.stop()
        self.profiler.print()

        proc: psutil.Process = psutil.Process(os.getpid())
        print(f"memory used: {bytes2human(proc.memory_info().rss)}")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Package definitions.

see copyright/license https://github.com/DerwenAI/strwythura/README.md
"""

from .ctx import DomainContext, TextChunk

from .elem import Entity, EntityInstance, EntitySource, NodeKind, NounSpan, \
    de_token_span, \
    STRW_BASE, STRW_PREFIX

from .ent import EntityStore

from .erkg import KnowledgeGraph

from .lex import LexicalGraph

from .nlp import Parser

from .opt import calc_quantile_bins, root_mean_square, stripe_column

from .prof import Profiler

from .rag import GraphRAG

from .resources import STRW_LOGO, SZ_LOGO

from .scrape import Scraper

from .vis import VisHTML

from .work import Workflow



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Manage the domain context.

see copyright/license https://github.com/DerwenAI/strwythura/README.md
"""

from collections import Counter, defaultdict
import itertools
import typing

from icecream import ic
from lancedb.embeddings import get_registry  # type: ignore
from lancedb.embeddings.sentence_transformers import SentenceTransformerEmbeddings  # type: ignore
from lancedb.pydantic import LanceModel, Vector  # type: ignore
from rdflib import Namespace  # pylint: disable=W0611
from rdflib.namespace import RDF
from rdflib.plugins.sparql.processor import SPARQLResult
from sz_semantics import Thesaurus  # type: ignore
import lancedb  # type: ignore
import polars as pl
import spacy

from .elem import Entity, EntitySource, NodeKind, NounSpan, \
    STRW_PREFIX
from .ent import EntityStore
from .erkg import KnowledgeGraph
from .lex import LexicalGraph


EMBED_FUNC: SentenceTransformerEmbeddings = \
    get_registry().get("sentence-transformers").create()


class TextChunk (LanceModel):
    """
Represents one chunk of text from a document.
    """
    uid: int
    url: str
    sent_id: int
    text: str = EMBED_FUNC.SourceField()
    vector: Vector(EMBED_FUNC.ndims()) = EMBED_FUNC.VectorField(default = None)  # type: ignore

    @classmethod
    def get_iri (
        cls,
        uid: int,
        ) -> str:
        """
Construct an IRI based on the chunk `uid` value.
        """
        return f"{STRW_PREFIX}chunk_{uid}"


    @classmethod
    def get_uid (
        cls,
        iri: str,
        ) -> int:
        """
Extract the `uid` value based on a chunk IRI.
        """
        stub: str = f"{STRW_PREFIX}chunk_"
        return int(iri.replace(stub, ""))


class DomainContext:
    """
Represent the domain context using an _ontology pipeline_ process:
vocabulary, taxonomy, thesaurus, and ontology.
    """
    TAXO_SENT_ID: int = 0


    def __init__ (
        self,
        config: dict,
        thesaurus: Thesaurus,
        ) -> None:
        """
Constructor.
        """
        self.config: dict = config

        # entities and semantic layer
        self.thesaurus: Thesaurus = thesaurus
        self.ent_store: EntityStore = EntityStore(self.config)

        # intermediate parsing outcomes
        self.lex: LexicalGraph = LexicalGraph(self.config)

        # the constructed ERKG knowledge graph in `NetworkX`
        self.erkg: KnowledgeGraph = KnowledgeGraph(self.config)

        # the vector store in `LanceDB`
        self.lancedb_conn: lancedb.db.LanceDBConnection = lancedb.connect(
            self.config["vect"]["lancedb_uri"],
        )

        self.start_chunk_id: int = 0
        self.chunk_table: lancedb.table.LanceTable | None = None


    ######################################################################
    ## manage the vector store

    def open_vector_tables (
        self,
        *,
        create: bool = False,
        ) -> None:
        """
Open the table for text chunk embeddings in the vector store,
overwriting any previous data if indicated.
        """
        if create:
            # intialize and clear any previous table
            self.chunk_table = self.lancedb_conn.create_table(
                self.config["vect"]["chunk_table"],
                schema = TextChunk,
                mode = "overwrite",
            )
        else:
            # open existing table
            self.chunk_table = self.lancedb_conn.open_table(
                self.config["vect"]["chunk_table"],
            )

            df_chunks: pl.DataFrame = self.chunk_table.search().select([ "uid" ]).to_polars()
            uids: list[ int ] = [ int(row[0]) for row in df_chunks.iter_rows() ]  # type: ignore

            if len(uids) > 0:
                self.start_chunk_id = max(uids) + 1
            else:
                self.start_chunk_id = 0


    def add_chunk (
        self,
        url: str,
        sent_id: int,
        text: str,
        ) -> TextChunk:
        """
Add a chunk into both the vector store and the ERKG.
        """
        chunk: TextChunk = TextChunk(
            uid = self.start_chunk_id,
            url = url,
            sent_id = sent_id,
            text = text,
        )

        ## add to the vector store
        self.chunk_table.add([ chunk ])  # type: ignore
        self.start_chunk_id += 1

        # add node to the ERKG
        self.erkg.add_node(
            TextChunk.get_iri(chunk.uid),
            NodeKind.CHUNK,
            attrs = {
                "chunk" : chunk.uid,
                "url": chunk.url,
                "sent": chunk.sent_id,
            },
        )

        return chunk


    def get_chunk_meta (
        self,
        ) -> typing.Iterator[ tuple[ int, str ]]:
        """
Iterator for TextChunk metadata from the `LanceDB` table.
        """
        for uid, url in self.chunk_table.search().select([ "uid", "url" ]).to_polars().iter_rows():  # type: ignore
            yield uid, url


    ######################################################################
    ## manage the semantics

    def get_label_map (
        self,
        ) -> dict[ str, str ]:
        """
Accessor: iterate through `skos:Concept` entities to extract a mapping
between NER labels and abbreviated IRIs.

Used for _entity linking_ in using zero-shot NER tasks, such as the `GLiNER`
library.
        """
        query: str = """
SELECT DISTINCT ?concept_iri ?label
WHERE {
  ?concept_iri a skos:Concept ;
    skos:prefLabel ?label ;
    sz:ner_label true ;
  .
}""".strip()

        qres: SPARQLResult = self.thesaurus.rdf_graph.query(query)

        return {
            row[1].toPython(): self.thesaurus.n3(row[0])  # type: ignore
            for row in qres
        }


    def promote_data_nodes (
        self,
        *,
        debug: bool = False,
        ) -> None:
        """
Reform selected RDF triples in `RDFlib` to be represented in a `NetworkX`
property graph: for the provenance of data records used in ER.
        """
        query: str = """
SELECT DISTINCT ?rec_iri ?rec_key ?data_src
WHERE {
  ?rec_iri a sz:DataRecord ;
    dc:identifier ?rec_key ;
    prov:wasQuotedFrom ?data_src ;
  .
}""".strip()

        # iterate through the SPARQL query results
        qres: SPARQLResult = self.thesaurus.rdf_graph.query(query)

        for row in qres:
            rec_iri: str = self.thesaurus.n3(row[0])  # type: ignore
            rec_key: str = row[1].toPython().strip()  # type: ignore
            data_src: str = self.thesaurus.n3(row[2])  # type: ignore

            if debug:
                ic(rec_iri, rec_key, data_src)

            # add node to the ERKG
            self.erkg.add_node(
                rec_iri,
                NodeKind.DATAREC,
                attrs = {
                    "rec_key": rec_key,
                    "data_src": data_src,
                },
            )


    def promote_taxo_nodes (
        self,
        *,
        debug: bool = False,
        ) -> None:
        """
Reform selected RDF triples in `RDFlib` to be represented in a `NetworkX`
property graph: for the `SKOS:Concept` items from the taxonomy.

Also add embeddings for each `SKOS:definition` text in the vector store.
        """
        query: str = """
SELECT DISTINCT ?concept_iri ?text ?lemma
WHERE {
  ?concept_iri a skos:Concept ;
    skos:definition ?text ;
    sz:lemma_phrase ?lemma ;
  .
}""".strip()

        qres: SPARQLResult = self.thesaurus.rdf_graph.query(query)

        # iterate through the SPARQL query results
        for i, row in enumerate(qres):
            concept_iri: str = self.thesaurus.n3(row[0])  # type: ignore
            text: str = row[1].toPython()  # type: ignore
            lemma_key: str = row[2].toPython()  # type: ignore

            if debug:
                ic(i, concept_iri, text, lemma_key)

            # create an entry in the entity store
            ent: Entity = Entity(
                span = NounSpan(
                    loc = ( -1, -1, ),
                    text = "",
                    span = [],
                    source = EntitySource.TAXO,
                ),
                lemma_key = lemma_key,
            )

            found_ent: Entity = self.ent_store.encode_entity(  # type: ignore
                ent,
                create = True,
            )

            # add the text and its embedding for the `SKOS:definition`
            # as a chunk in the vector store
            chunk_id: int = self.add_chunk(  # pylint: disable=W0612
                concept_iri,
                self.TAXO_SENT_ID, # zero sentence reserved for taxonomy concepts
                text,
            )

            # add node to the ERKG
            self.erkg.add_node(
                concept_iri,
                NodeKind.TAXONOMY,
                attrs = {
                    "count": found_ent.count,
                    "rank": found_ent.rank,
                    "text": text,
                    "lemma": lemma_key,
                    "method": EntitySource.TAXO.value,
                },
                stop = False,
            )

        # query for SKOS relations within the taxonomy,
        # then add ERKG edges to represent these
        query = """
SELECT DISTINCT ?ent ?sem_rel ?rel
WHERE {
  VALUES ?sem_rel {
    skos:broader
    skos:narrower
    skos:related
  } .
  ?ent a skos:Concept ;
    ?sem_rel ?rel .
}""".strip()

        qres = self.thesaurus.rdf_graph.query(query)

        # iterate through the SPARQL query results
        for row in qres:
            ent_iri: str = self.thesaurus.n3(row[0])  # type: ignore
            sem_rel: str = self.thesaurus.n3(row[1])  # type: ignore
            rel_iri: str = self.thesaurus.n3(row[2])  # type: ignore

            if debug:
                ic(ent_iri, sem_rel, rel_iri)

            # add a ERKG edge for the related entities
            self.erkg.add_edge(
                ent_iri,
                sem_rel,
                rel_iri,
	        prob = 1.0,
                update = True,
            )


    def promote_er_nodes (
        self,
        parser: "Parser",  # type: ignore
        *,
        debug: bool = False,
        ) -> None:
        """
Reform selected RDF triples in `RDFlib` to be represented in a `NetworkX`
property graph: for the entity definitions from ER.
        """
        query: str = """
SELECT DISTINCT ?ent ?ent_class ?label
WHERE {
  VALUES ?ent_class {
    sz:Person
    sz:Organization
  } .
  ?ent a ?ent_class ;
    skos:prefLabel ?label ;
  .
}""".strip()

        # iterate through the SPARQL query results
        qres: SPARQLResult = self.thesaurus.rdf_graph.query(query)

        for row in qres:
            ent_iri: str = self.thesaurus.n3(row[0])  # type: ignore
            concept_iri: str = self.thesaurus.n3(row[1])  # type: ignore
            label: str = row[2].toPython().strip()  # type: ignore
            rank: float = 0.0

            if debug:
                ic(ent_iri, concept_iri, label)

            ## TODO: THIS LOGIC IS HORKED!!
            ## each ER needs to become a distinct entity
            if len(label) < 1:
                # create a ERKG node, though without an entity definition
                #uid: int = self.ent_store.increment_uid()
                lemma_key: str = ""
                label = ent_iri

            else:
                # use the label to generate a lemma key
                span: spacy.tokens.doc.Doc = parser.ner_pipe(label)
                lemma_key = parser.tokenize_lemma(span)

                # create an entry in the entity store
                ent: Entity = Entity(
                    span = NounSpan(
                        loc = ( 0, len(span) - 1, ),
                        text = label,
                        span = span,  # type: ignore
                        source = EntitySource.ER,
                        iri = ent_iri,
                    ),
                    lemma_key = lemma_key,
                )

                found_ent: Entity = self.ent_store.encode_entity(  # type: ignore
                    ent,
                    create = True,
                )

                rank = found_ent.rank

            # add node to the ERKG
            self.erkg.add_node(
                ent_iri,
                NodeKind.ENTITY,
                attrs = {
                    "count": found_ent.count,
                    "rank": rank,
                    "text": label,
                    "lemma": lemma_key,
                    "method": EntitySource.ER.value,
                }
            )

            # add a ERKG edge to link to the SKOS:concept class
            self.erkg.add_edge(
	        ent_iri,
                self.thesaurus.n3(RDF.type),
                concept_iri,
	        prob = 1.0,
                update = True,
            )


    def promote_er_edges (
        self,
        *,
        debug: bool = False,
        ) -> None:
        """
Reform selected RDF triples in `RDFlib` to be represented in a `NetworkX`
property graph: for the SKOS relations from ER.
        """
        # query blank nodes for ent => ent | rec
        # SKOS relations, then add edges
        query: str = """
SELECT DISTINCT ?ent ?rel_ent ?sem_rel ?key ?lev
WHERE {
  ?bl rdf:predicate ?sem_rel ;
    rdf:subject ?ent ;
    rdf:object ?rel_ent ;
    sz:match_key ?key ;
    sz:match_level ?lev ;
  .
}""".strip()

        # iterate through the SPARQL query results
        qres: SPARQLResult = self.thesaurus.rdf_graph.query(query)

        for row in qres:
            ent_iri: str = self.thesaurus.n3(row[0])  # type: ignore
            rel_iri: str = self.thesaurus.n3(row[1])  # type: ignore

            sem_rel: str = self.thesaurus.n3(row[2])  # type: ignore
            prob: float = 1.0

            match_key: str = row[3].toPython().strip()  # type: ignore
            match_level: int = row[4].toPython()  # type: ignore

            match match_level:
                case 11:
                    # Senzing calls this a `disclosed relationship`
                    # which alternatively might be `sz:member_of`
                    prob = 1.0
                case 2:
                    prob = 0.8
                case 3:
                    prob = 0.5

            if debug:
                ic(ent_iri, rel_iri, sem_rel, match_key, match_level)

            # add a ERKG edge for ent => ent | rec relations
            self.erkg.add_edge(
	        ent_iri,
                sem_rel,
                rel_iri,
	        prob = prob,
                attrs = {
                    "match_key": match_key,
                    "match_level": match_level,
                },
                update = True,
            )


    def promote_ner_nodes (
        self,
        *,
        debug: bool = False,  # pylint: disable=W0613
        ) -> None:
        """
Reform selected RDF triples in `RDFlib` to be represented in a `NetworkX`
property graph: for the entities extracted from NER.
        """
        # iterate through the entity store
        for ent in self.ent_store.entities.values():
            if ent.span.source >= EntitySource.NER:
                # add node to the ERKG
                self.erkg.add_node(
                    ent.get_iri(),
                    NodeKind.ENTITY,
                    attrs = {
                        "count": ent.count,
                        "rank": ent.rank,
                        "text": ent.span.text,
                        "lemma": ent.lemma_key,
                        "method": ent.span.source.value,
                    },
                    update = True,
                )

                # add a ERKG edge to link to the SKOS:concept class
                self.erkg.add_edge(
	            ent.get_iri(),
                    self.thesaurus.n3(RDF.type),
                    ent.span.iri,
	            prob = 1.0,
                    update = True,
                )


    ######################################################################
    ## manage additional entity context

    def link_entity_chunks (
        self,
        *,
        debug: bool = False,  # pylint: disable=W0613
        ) -> None:
        """
Cross-link entities with the chunks in which they appear.
        """
        sem_rel: str = f"{STRW_PREFIX}within_chunk"

        for ent in self.ent_store.entities.values():
            for ent_inst in ent.inst:
                self.erkg.add_edge(
                    ent.get_iri(),
                    sem_rel,
                    TextChunk.get_iri(ent_inst.chunk_id),
                    prob = 1.0,
                    attrs = {
                        "weight": ent.rank,
                        "sent": ent_inst.sent_id,
                    },
                    update = True,
                )


    def co_occur_entities (
        self,
        *,
        debug: bool = False,
        ) -> None:
        """
Connect entities which co-occur within the same sentence.
        """
        inst_dict: dict[ int, dict[ int, list[ int ]]] = defaultdict(lambda: defaultdict(list))
        counter: Counter = Counter()

        # partition entity co-occurrence by `( chunk_id, sent_id, ent.uid, )`
        for ent in self.ent_store.entities.values():
            if ent.span.source >= EntitySource.NER:
                for ent_inst in ent.inst:
                    inst_dict[ent_inst.chunk_id][ent_inst.sent_id].append(ent.uid)

        if debug:
            ic(inst_dict)

        # tally the pairwise co-occurrence of entities
        for _, sent_dict in sorted(inst_dict.items()):
            for _, ent_list in sorted(sent_dict.items()):
                for pair in itertools.combinations(ent_list, 2):  # type: ignore
                    pair = tuple(sorted(pair))  # type: ignore
                    counter[pair] += 1

                    pair = tuple(sorted(pair, reverse = True))  # type: ignore
                    counter[pair] += 1

        if debug:
            ic(counter)

        # partition by first element, to compute a conditional
        # probability per second element
        tally: dict[ int, list ] = defaultdict(list)
        decoder: dict[ int, Entity ] = self.ent_store.get_decoder()
        sem_rel: str = f"{STRW_PREFIX}co_occurs_with"

        for pair, count in counter.items():
            tally[pair[0]].append(( pair, count, ))

        for _, pairs_list in tally.items():
            partition: Counter = Counter(dict(pairs_list))
            total: float = float(partition.total())

            for pair, count in partition.items():
                prob: float = round(float(count) / total, 4)

                if debug:
                    ic(pair, count, prob)

                # add relation into the ERKG
                self.erkg.add_edge(
                    decoder[pair[0]].get_iri(),
                    sem_rel,
                    decoder[pair[1]].get_iri(),
                    prob = prob,
                    update = True,
                )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Manage the ERKG knowledge graph, based on `NetworkX`.

see copyright/license https://github.com/DerwenAI/strwythura/README.md
"""

import inspect
import json
import math
import pathlib
import sys
import typing

from icecream import ic
import networkx as nx

from .elem import EntitySource, NodeKind


class KnowledgeGraph:
    """
Represent the ERKG knowledge graph asset.
    """

    def __init__ (
        self,
        config: dict,
        ) -> None:
        """
Constructor.
        """
        self.config: dict = config
        self.graph: nx.MultiDiGraph = nx.MultiDiGraph()


    ######################################################################
    ## serialization

    def load_graph (
        self,
        erkg_path: pathlib.Path,
        ) -> None:
        """
De-serialize a constructed KG from a JSON file represented in the
_node-link_ data format.
        """
        with erkg_path.open("r", encoding = "utf-8") as fp:
            self.graph = nx.node_link_graph(
                json.load(fp),
                edges = "edges",
            )


    def save_graph (
        self,
        erkg_path: pathlib.Path,
        ) -> None:
        """
Serialize the constructed KG as a JSON file represented in the
_node-link_ data format.

Aternatively this could be stored in a graph database.
        """
        with erkg_path.open("w", encoding = "utf-8") as fp:
            fp.write(
                json.dumps(
                    nx.node_link_data(
                        self.graph,
                        edges = "edges",
                    ),
                    indent = 2,
                    sort_keys = True,
                )
            )


    ######################################################################
    ## graph construction

    def add_edge (  # pylint: disable=W0102
        self,
        src_iri: str,
        rel_iri: str,
        dst_iri: str,
        prob = 0.0,
        *,
        attrs: dict = {},
        update: bool = False,
        stop: bool = True,
        debug: bool = False,
        ) -> dict | None:
        """
Add an edge into the ERKG with required and optional properties.

Required properties: each edge must have an IRI as its `MultiGraph`
unique key, and a `prob` probability value.

Optional properties: specified as key/value pairs in the `attrs`
dictionary.
        """
        edge: tuple = ( src_iri, dst_iri, rel_iri, )
        pre_exist: bool = False

        # override conflicting settings
        if update:
            stop = False

        # test whether the edge IRI already exists in the ERKG?
        if self.graph.has_edge(*edge):
            pre_exist = True

            calframe: list = inspect.getouterframes(inspect.currentframe(), 2)
            caller: str = calframe[1][3]
            prev_attrs: dict = self.graph.edges[*edge]

            if debug | stop:
                print(f"dupe: {caller} {edge} {prob} {attrs}")
                print("PRE-EXISTING EDGE", prev_attrs)

            if stop:
                # if requested for debugging, stop the application
                sys.exit(-1)
            elif not update:
                # return the pre-existing edge data and do not update
                return prev_attrs

        # add an edge into the ERKG
        if not pre_exist:
            if debug:
                ic("ADD EDGE", edge, prob, attrs)

            self.graph.add_edge(
	        src_iri,
                dst_iri,
	        key = rel_iri,
	        prob = prob,
            )

        # set the optional edge attributes, if any
        if (update or not pre_exist) and len(attrs) > 0:
            nx.set_edge_attributes(
                self.graph,
                { edge: attrs },
            )

        return None


    def add_node (  # pylint: disable=W0102
        self,
        iri: str,
        kind: NodeKind,
        *,
        attrs: dict = {},
        update: bool = False,
        stop: bool = True,
        debug: bool = False,
        ) -> dict | None:
        """
Add a node into the ERKG with required and optional properties.

Required properties: each node must have an IRI as its unique
identifier, and a `NodeKind` value.

Optional properties: specified as key/value pairs in the `attrs`
dictionary.
        """
        pre_exist: bool = False

        # override conflicting settings
        if update:
            stop = False

        # test whether the node IRI already exists in the ERKG?
        if self.graph.has_node(iri):
            pre_exist = True

            calframe: list = inspect.getouterframes(inspect.currentframe(), 2)
            caller: str = calframe[1][3]
            prev_attrs: dict = self.graph.nodes[iri]

            if debug | stop:
                print(f"dupe: {caller} {iri} {kind}")
                print("PRE-EXISTING NODE", prev_attrs)

            if stop:
                # if requested for debugging, stop the application
                sys.exit(-1)
            elif not update:
                # return the pre-existing node data and do not update
                return prev_attrs

        # add a node into the ERKG
        if not pre_exist:
            if debug:
                ic("ADD NODE", iri, kind.value, attrs)

            self.graph.add_node(
                iri,
                kind = kind.value,
            )
        else:
            attrs["kind"] = kind.value

        # set the optional node attributes, if any
        if (update or not pre_exist) and len(attrs) > 0:

            nx.set_node_attributes(
                self.graph,
                { iri: attrs },
            )

        return None


    ######################################################################
    ## accessors for graph elements and views

    def get_node (
        self,
        iri: str,
        ) -> dict:
        """
Accessor method to get the properties of an ERKG node.
        """
        return self.graph.nodes[iri]


    def inbound_edges (
        self,
        dst_iri: str,
        *,
        attr: str = "weight",
        ) -> typing.Iterator[ tuple[ str, set[ str ], float ] ]:
        """
Iterator for the inbound connections of an ERKG node.
        """
        for src_iri, _, keys, val in self.graph.in_edges(  # type: ignore
            nbunch = dst_iri,
            data = attr,
            keys = True,
        ):
            yield src_iri, keys, val


    def neighbors (
        self,
        iri: str,
        ) -> typing.Iterator[ str ]:
        """
Iterator for the neighbors of an ERKG node.
        """
        return self.graph.neighbors(iri)


    def subgraph (
        self,
        iri_set: set[ str ],
        ) -> nx.Graph:
        """
Returns a subgraph view from the given node IRIs.
        """
        return self.graph.subgraph(iri_set)


    def shortest_paths (
        self,
        src_iri: str,
        dst_iri: str,
        ) -> typing.Iterator[ list[ str ] ]:
        """
Compute all shortest simple paths in the graph.
        """
        return nx.all_shortest_paths(self.graph, src_iri, dst_iri)


    ######################################################################
    ## interactive visualization

    def vis_nodes (
        self,
        num_docs: int,
        ) -> typing.Iterator[ tuple[ int, dict, ] ]:
        """
Iterator for the visualization attributes of nodes.
        """
        filter_lex: set[ str ] = set([ EntitySource.NC.value, EntitySource.LEX.value ])

        for iri, attrs in self.graph.nodes(data = True):
            attr: dict = {}

            if attrs.get("kind") == NodeKind.ENTITY.value and attrs.get("label") not in filter_lex:
                attr["color"] = "hsla(65, 46%, 58%, 0.80)"
                attr["size"] = round(20 * math.log(1.0 + math.sqrt(float(attrs.get("count"))) / num_docs))
                attr["label"] = attrs.get("text")  # type: ignore
                attr["title"] = attrs.get("key")  # type: ignore
            elif attrs.get("kind") == NodeKind.TAXONOMY.value:
                attr["color"] = "hsla(306, 45%, 57%, 0.95)"
                attr["size"] = 5
                attr["label"] = attrs.get("label")  # type: ignore
                attr["title"] = attrs.get("iri")  # type: ignore
            else:
                continue

            yield iri, attr


    def vis_edges (
        self,
        ) -> typing.Iterator[ tuple[ int, int, str, ] ]:
        """
Iterator for the visualization attributes of edges.
        """
        for src_iri, dst_iri, key in self.graph.edges(keys = True):
            yield src_iri, dst_iri, key



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Package definitions for the resource files.

see copyright/license https://github.com/DerwenAI/strwythura/README.md
"""

import pathlib

STRW_LOGO: pathlib.Path = pathlib.Path(__file__).resolve().parent / "logo.png"
SZ_LOGO: pathlib.Path = pathlib.Path(__file__).resolve().parent / "senzing.png"



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Manage the natural language processing.

see copyright/license https://github.com/DerwenAI/strwythura/README.md
"""

import typing
import warnings

from icecream import ic
import spacy

from .ctx import DomainContext
from .elem import Entity, EntityInstance, EntitySource, NounSpan


class Parser:
    """
Wrapper class for the `spaCy` NLP pipeline used to extract entities
and relations, based on `GLiNER`, `DSPy`, and textgraphs.

The default definitions are for American English usage. Subclass
to modify the definitions for other languages and dialects.
    """
    BASE_CONCEPT: str = "owl:Thing"

    STOP_WORDS: set[ str ] = set([
        "PRON.each",
        "PRON.he",
        "PRON.it",
        "PRON.she",
        "PRON.some",
        "PRON.someone",
        "PRON.that",
        "PRON.their",
        "PRON.they",
        "PRON.those",
        "PRON.we",
        "PRON.what",
        "PRON.which",
        "PRON.who",
        "PRON.you",
    ])

    POS_TRANSFORM: dict[ str, str ] = {
        "PROPN": "NOUN",
    }


    def __init__ (
        self,
        config: dict,
        ctx: DomainContext,
        ) -> None:
        """
Constructor.
        """
        self.config: dict = config
        self.ctx: DomainContext = ctx
        self.label_map: dict[ str, str ] = self.ctx.get_label_map()
        self.ner_pipe: spacy.Language = self.build_ner_pipe()


    def build_ner_pipe (
        self,
        *,
        use_gliner: bool = True, # False
        ) -> spacy.Language:
        """
Initialize the `spaCy` pipeline used for NER + RE, by loading models
for `spaCy`, `GLiNER`

  - `ner_labels`: semantics to apply for zero-shot NER

This assumes the `spaCy` model has been downloaded already.

Note: this may take several minutes when run the first time after
installing the repo.
        """
        ner_pipe: spacy.Language = spacy.load(self.config["nlp"]["spacy_model"])

        if use_gliner:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")

                ner_pipe.add_pipe(
                    "gliner_spacy",
                    config = {
                        "style": "ent",
                        "labels": list(self.label_map.keys()),
                        "gliner_model": self.config["nlp"]["gliner_model"],
                        "chunk_size": self.config["vect"]["chunk_size"],
                    },
                )

        return ner_pipe


    def normalize_pos (
        self,
        pos: str,
        ) -> str:
        """
Normalize the _part-of-speech_ for downstream usage.
        """
        if pos in self.POS_TRANSFORM:
            return self.POS_TRANSFORM[pos]

        return pos


    def tokenize_lemma (
        self,
        span: list,
        *,
        debug: bool = False,
        ) -> str:
        """
Construct a parsed, lemmatized key for the given noun phrase.
        """
        lemma_key: str = " ".join([
            f"{self.normalize_pos(tok.pos_)}.{tok.lemma_.strip().lower()}"
            for tok in span
        ])

        if debug:
            ic(lemma_key, span)

        return lemma_key


    @classmethod
    def within_loc (
        cls,
        loc0: tuple[ int, int ],
        loc1: tuple[ int, int ],
        ) -> bool:
        """
Test whether the `loc0` tuple is within the `loc1` tuple.
        """
        return loc0[0] >= loc1[0] and loc0[1] <= loc1[1]

    @classmethod
    def overlaps_loc (
        cls,
        loc0: tuple[ int, int ],
        loc1: tuple[ int, int ],
        ) -> bool:
        """
Test whether the `loc0` tuple overlaps the `loc1` tuple.
        """
        if loc0[0] == loc0[1] or loc1[0] == loc1[1]:
            return False

        head: bool = loc0[0] >= loc1[0] and loc0[0] <= loc1[1]
        tail: bool = loc0[1] >= loc1[0] and loc0[1] <= loc1[1]

        return head ^ tail


    def transform_sentence (
        self,
        sent: spacy.tokens.span.Span,
        *,
        debug: bool = False,
        ) -> typing.Iterator[ NounSpan ]:
        """
Transform a parsed sentence into a sequence of the following, in order
of priority:

  * NER spans
  * noun chunks
  * tokens
        """
        nck_list: list[ NounSpan ] = [
            NounSpan(
                loc = ( nck.start, nck.end - 1, ),
                text = nck.text,
                span = nck,  # type: ignore
                source = EntitySource.NC,
                iri = self.BASE_CONCEPT,
            )
            for nck in sent.noun_chunks
            if len(nck) > 1
        ]

        if debug:
            ic(nck_list)

        ner_list: list[ NounSpan ] = [
            NounSpan(
                loc = ( ner.start, ner.end - 1, ),
                text = ner.text,
                span = ner,  # type: ignore
                label = ner.label_,
                source = EntitySource.NER,
                iri = self.label_map[ner.label_],
            )
            for ner in sent.ents
        ]

        if debug:
            ic(ner_list)

        tok_list: list[ NounSpan ] = [
            NounSpan(
                loc = ( tok.i, tok.i, ),
                text = tok.text.strip(),
                span = [ tok ],
                label = self.normalize_pos(tok.pos_),
                iri = self.BASE_CONCEPT,
            )
            for tok in sent
        ]

        # consolidate the noun chunks with named entities,
        # producing a list of spans
        span_list: list[ NounSpan ] = []

        todo: tuple | None = None

        while len(nck_list) > 0 and len(ner_list) > 0:
            last_todo: tuple | None = todo
            todo = ( len(nck_list), len(ner_list), )

            if debug:
                ic(todo, last_todo)

            if todo == last_todo:
                print("NCK list", nck_list)
                print("NER list", ner_list)
                raise RuntimeError("infinite loop")

            if debug:
                print("todo", todo, sent)

            nck: NounSpan = nck_list[0]
            ner: NounSpan = ner_list[0]

            if debug:
                ic(nck.loc, ner.loc)
                ic(self.within_loc(nck.loc, ner.loc))
                ic(self.overlaps_loc(nck.loc, ner.loc))

            #if nck.loc[1] == ner.loc[0]:
            if self.overlaps_loc(nck.loc, ner.loc):
                # NCK overlaps an NER
                nck_list.pop(0)

                ner_list.pop(0)
                span_list.append(ner)

            elif self.within_loc(nck.loc, ner.loc):
                # NER subsumes NCK
                if debug:
                    print("NER subsumes NCK", ner)

                while len(nck_list) > 0 and ner.loc[0] <= nck_list[0].loc[0] and ner.loc[1] >= nck_list[0].loc[1]:
                    nck_list.pop(0)

                ner_list.pop(0)
                span_list.append(ner)

            elif ner.loc[1] < nck.loc[0]:
                # emit NER
                if debug:
                    print("emit NER", ner)

                ner_list.pop(0)
                span_list.append(ner)

            elif ner.loc[0] > nck.loc[1]:
                # emit unlabeled NC
                if debug:
                    print("emit NCK", nck)

                nck_list.pop(0)
                span_list.append(nck)

            elif self.within_loc(ner.loc, nck.loc):
                # NCK subsumes NER, it inherits the label, iri, etc.
                if debug:
                    print("NCK subsumes NER", nck)

                ner_list.pop(0)
                nck.source = ner.source

                if nck.label is None:
                    nck.label = ner.label

                if nck.iri is None:
                    nck.iri = ner.iri

        span_list.extend(nck_list)
        span_list.extend(ner_list)
        span_list.sort(key = lambda span: span.loc)

        # consolidate the span list with tokens, producing a full list of parsed items
        full_list: list[ NounSpan ] = []

        while len(span_list) > 0 and len(tok_list) > 0:
            span: NounSpan = span_list[0]
            tok: NounSpan = tok_list[0]

            if tok.loc[0] < span.loc[0]:
                # emit token
                tok_list.pop(0)
                full_list.append(tok)

            elif tok.loc[0] >= span.loc[0] and tok.loc[0] <= span.loc[1]:
                # span subsumes token
                tok_list.pop(0)

            elif tok.loc[0] > span.loc[1]:
                # emit span
                span_list.pop(0)
                full_list.append(span)

        full_list.extend(span_list)
        full_list.extend(tok_list)
        full_list.sort(key = lambda span: span.loc)

        yield from full_list


    def parse_para (
        self,
        chunk_id: int,
        chunk_text: str,
        *,
        debug: bool = False,
        ) -> int:
        """
Parse a text paragraph, then per sentence:

  * transform into a sequence of NER spans, noun chunks, or tokens
  * extract entities, with labels mapped to IRIs where possible
  * tokenize entities based on lemmatization
  * use `textgraph` to add to the `LexicalGraph` lexical graph

For the paragraph, load an entity sequence vector into `gensim.Word2Vec`
using `EntityStore`
        """
        doc: spacy.tokens.doc.Doc = self.ner_pipe(chunk_text)
        num_sent: int = 0
        ent_seq: list[ Entity ] = []

        # transform sentence as: NER spans, noun chunks, tokens
        for sent_id, sent in enumerate(doc.sents):
            num_sent = sent_id
            sent_text: str = str(sent).strip()

            if debug:
                ic(sent_id, sent_text)

            ent_inst: EntityInstance = EntityInstance(
                sent_id = sent_id,
                chunk_id = chunk_id,
            )

            if debug:
                ic(ent_inst)

            for item in self.transform_sentence(sent, debug = debug):
                if debug:
                    ic(ent_seq, item)

                if item.source == EntitySource.LEX and item.label != "NOUN":
                    continue

                # tokenize entities based on lemmatization
                lemma_key: str = self.tokenize_lemma(item.span)

                if lemma_key in self.STOP_WORDS:
                    continue

                ent: Entity = Entity(
                    span = item,
                    lemma_key = lemma_key,
                )

                found_ent: Entity = self.ctx.ent_store.encode_entity(  # type: ignore
                    ent,
                    create = True,
                )

                found_ent.inst.append(ent_inst)
                ## TODO: make sequence based on paragraph
                ent_seq.append(found_ent)

            if debug:
                ic(sent_id, sent, ent_seq)

            # add this sentence to the _textgraph_ in `LexicalGraph`
            self.ctx.lex.add_sent(ent_seq)

        # load an entity sequence vector into `gensim.Word2Vec` using `EntityStore`
        seq_vec: list[ int ] = [
            ent.uid
            for ent in ent_seq
            if ent.uid is not None
        ]

        self.ctx.ent_store.embed_sequence(seq_vec)

        if debug:
            print(seq_vec)

        return num_sent + 1



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Run an enhanced GraphRAG based on ERKG and entity embedding, using DSPy.

see copyright/license https://github.com/DerwenAI/strwythura/README.md
"""

import itertools
import os
import time
import traceback
import typing

from datasketch import MinHashLSHForest, MinHash  # type: ignore
from icecream import ic
from opik.integrations.dspy.callback import OpikCallback
import dspy  # type: ignore
import networkx as nx
import opik
import spacy

from .ctx import TextChunk
from .elem import Entity, EntitySource, NodeKind, STRW_PREFIX
from .work import Workflow


class DSPy_RAG (dspy.Module):  # pylint: disable=C0103
    """
DSPy implementation of a RAG signature.
    """

    def __init__(  # pylint: disable=W0231
        self,
        config: dict,
        project_name: str,
        *,
        run_local: bool = True,
        use_opik: bool = True,
        ) -> None:
        """
Constructor.
        """
        self.config: dict = config

        # load the LLM
        dspy.configure(
            track_usage = True,
        )

        if run_local:
            self.lm: dspy.LM = dspy.LM(
                self.config["rag"]["lm_name"],
                api_base = self.config["rag"]["api_base"],
                api_key = "",
                temperature = self.config["rag"]["temperature"],
                max_tokens = self.config["rag"]["max_tokens"],
                stop = None,
                cache = False,
            )
        else:
            OPENAI_API_KEY: str = os.environ.get("OPENAI_API_KEY")  # type: ignore  # pylint: disable=C0103

            if OPENAI_API_KEY is None:
                raise ValueError(
                    "Environment variable 'OPENAI_API_KEY' is not set. Please set it to proceed."
                )

            self.lm = dspy.LM(
                "openai/gpt-4o-mini",
                temperature = 0.0,
            )

        # set up `Opik` for observability
        # see: https://www.comet.com/docs/opik/python-sdk-reference/configure.html
        callbacks: list = []

        if use_opik:
            os.environ["OPIK_BASE_URL"] = self.config["opik"]["base_url"]

            opik.configure(
                use_local = True,
                url = self.config["opik"]["base_url"],
            )

            self.opik_callback: OpikCallback = OpikCallback(
                project_name = project_name,
                log_graph = True,
            )

            callbacks.append(self.opik_callback)

        # set up the `DSPy` signature for RAG
        dspy.configure(
            lm = self.lm,
            callbacks = callbacks,
        )

        # define the basic RAG signature
        self.respond: dspy.Predict = dspy.Predict(
            "context, question -> response"
        )

        self.context: list[ str ] = []


    def forward (
        self,
        question: str,
        ) -> dspy.primitives.prediction.Prediction:
        """
Invoke the RAG signature.
        """
        reply: dspy.primitives.prediction.Prediction = self.respond(
            context = self.context,
            question = question,
        )

        return reply


class GraphRAG:
    """
Run a question through `LanceDB` to identify related _chunks_ and
through the `Word2Vec` entity embedding model for _semantic expansion_
to produce a set of _anchor nodes_ in the `NetworkX` ERKG graph.
    """

    def __init__ (
        self,
        work: Workflow,
        project_name: str,
        project_description: str,
        *,
        run_local: bool = True,
        use_opik: bool = True,
        ) -> None:
        """
Constructor.
        """
        self.work: Workflow = work

        self.rag: DSPy_RAG = DSPy_RAG(
            self.work.config,
            project_name,
            run_local = run_local,
            use_opik = use_opik,
        )

        # search assets
        self.description: str = project_description
        self.anchor_nodes: set[ str ] = set()
        self.rag_chunks: dict[ int, float ] = {}
        self.subgraph: nx.Graph = nx.Graph()


    ######################################################################
    ## question/answer

    def question_answer (
        self,
        *,
        debug: bool = False,
        ) -> None:
        """
Loop to answer questions.
        """
        try:
            # loop to answer questions
            while True:
                question: str = input("\nQuoi? ").strip()

                if len(question) < 1:
                    continue

                if question.lower() in [ "quitter", "bye", "adieu" ]:
                    break

                # enchanced GraphRAG prioritizes and retrieves text chunks
                self.run_errag(
                    question,
                    debug = debug,
                )

                # LLM summarizes the text chunks in response to the question
                response: dspy.primitives.prediction.Prediction = self.qa_signature(
                    question,
                    self.get_chunks_text(),
                )

                ic(question)
                ic(response.response)
                ic(self.rag)
                print("-" * 10)

        except EOFError:
            print("")
        except Exception as ex:  # pylint: disable=W0718
            ic(ex)
            traceback.print_exc()
        finally:
            print("\nÀ bientôt!\n")
            time.sleep(.1)


    def qa_signature (
        self,
        question: str,
        chunks: list[ str ],
        ) -> dspy.primitives.prediction.Prediction:
        """
Run one question/answer cycle.
        """
        self.rag.context = chunks
        response: dspy.primitives.prediction.Prediction = self.rag(question)

        if False: # disable for now; too verbose  # pylint: disable=W0125
            dspy.inspect_history()

        return response


    ######################################################################
    ## enhanced GraphRAG methods

    def run_errag (
        self,
        question: str,
        *,
        disable_graph: bool = False,
        debug: bool = True, # False
        ) -> None:
        """
Run an enchanced GraphRAG to retrieve and prioritize text chunks
by leveraging the ERKG and entity embeddings.
        """
        max_chunks: int = self.work.config["rag"]["max_chunks"]
        num_perm: int = self.work.config["rag"]["num_perm"]

        # find the text chunks which are nearest to the question,
        # then find the entity nodes linked to these chunks
        chunk_nodes: dict[ str, float ] = self.find_rag_chunks(
            question,
            max_chunks = max_chunks,
        )

        if disable_graph:
            # disable the GraphRAG aspects, using RAG-only --
            # for testing and evaluation purposes
            return

        # find entities in the neighborhood of the question,
        # identifying the initial set of anchor nodes, plus MinHash
        # digests for the lemmatized phrases among these entities
        ner_mh: list[ MinHash ] = self.find_nearby_entities(
            question,
            num_perm = num_perm,
        )

        # use a locality-sensitive hash to filter the entity nodes
        # linked to chunks, to augment the set of anchor nodes
        self.augment_anchor_nodes(
            chunk_nodes,
            ner_mh,
            num_perm = num_perm,
        )

        # purely for debugging
        if debug:
            ic(self.rag_chunks)

            for iri in self.anchor_nodes:
                anchor: dict = self.work.ctx.erkg.get_node(iri)
                ic(anchor)

        # perform a semantic expansion using entity embeddings
        self.perform_semantic_expansion(
            w2v_top_k = self.work.config["rag"]["w2v_top_k"],
            w2v_min_dist = self.work.config["rag"]["w2v_min_dist"],
        )

        # extract a subgraph constructed from the shortest paths
        # between anchor nodes
        # TODO: if this variable isn't used, let's refactor it out
        try:
            subgraph: set[ str ] = set(list(self.extract_question_subgraph()))  # pylint: disable=W0612
        except Exception as ex:  # pylint: disable=W0718
            ic(ex)

        # add the chunks for each anchor node
        for chunk_id in self.find_chunk_neighbors():
            if chunk_id not in self.rag_chunks:
                # impute to median distance
                self.rag_chunks[chunk_id] = 0.5


    def find_rag_chunks (
        self,
        question: str,
        *,
        max_chunks: int = 11,
        debug: bool = False,
        ) -> dict[ str, float ]:
        """
Search the vector store for text chunks in the neighborhood of the
`question` prompt, which is the basic RAG process.

Then find graph nodes for entities linked to the selected chunks,
which are returned as a dictionary.
        """
        chunk_list: list[ dict ] = self.work.ctx.chunk_table.search(  # type: ignore
            question
        ).select(
            [ "uid", "_distance" ]
        ).limit(
            max_chunks
        ).to_list()

        chunk_nodes: dict[ str, float ] = {}
        sem_rel: str = f"{STRW_PREFIX}within_chunk"

        for row in chunk_list:
            chunk_id: int = int(row["uid"])
            chunk_iri: str = TextChunk.get_iri(chunk_id)
            distance: float = round((100.0 - row["_distance"]) / 100.0, 4)

            if debug:
                ic(chunk_iri, distance)

            self.rag_chunks[chunk_id] = distance

            for ent_iri, keys, val in self.work.ctx.erkg.inbound_edges(chunk_iri, attr = "weight"):  # type: ignore
                if sem_rel in keys:
                    metric: float = round(val * distance, 4)

                    if ent_iri not in chunk_nodes:
                        chunk_nodes[ent_iri] = metric
                    else:
                        chunk_nodes[ent_iri] = max(chunk_nodes[ent_iri], metric)

        return chunk_nodes


    def find_nearby_entities (
        self,
        question: str,
        *,
        num_perm: int = 128,
        debug: bool = False, # True
        ) -> list[ MinHash ]:
        """
Search the entity store for direct matches from NER
        """
        lem_seq: list[ str ] = []
        doc: spacy.tokens.doc.Doc = self.work.parser.ner_pipe(question)  # type: ignore

        lemma_todo: set[ str ] = {
            self.work.parser.tokenize_lemma(span)  # type: ignore
            for span in doc.ents
        }

        for sent in doc.sents:
            for item in self.work.parser.transform_sentence(sent):  # type: ignore
                lemma_key: str = self.work.parser.tokenize_lemma(item.span)  # type: ignore
                lem_seq.append(lemma_key)

                if debug:
                    ic(item, lemma_key)

                # try to find known entities directly
                # NB: this only works for single-word phrases
                if item.label in [ "ADP", "NOUN" ] and lemma_key not in self.work.parser.STOP_WORDS:  # type: ignore
                    found_ent: Entity = self.work.ctx.ent_store.encode_entity(  # type: ignore
                        Entity(span = item, lemma_key = lemma_key)
                    )

                    if found_ent is not None:
                        self.anchor_nodes.add(found_ent.get_iri())

                # as a fallback, keep the lemma keys for each noun phrase to use in an LSH
                if item.source in [ EntitySource.NER ]:
                    lemma_todo.add(lemma_key)

        if debug:
            ic(lemma_todo)
            ic(lem_seq)
            ic(self.anchor_nodes)

        # prepare for min hash approximation searches on lemmas
        ner_mh: list[ MinHash ] = []

        # the first hash is special: built on a sequence of lemma keys --
        # one per parsed token -- for the entire sentence
        ner_mh.append(MinHash(num_perm = num_perm))

        for lemma_key in lem_seq:
            for lemma in lemma_key.split(" "):
                ner_mh[0].update(lemma.encode("utf-8"))

                if debug:
                    ic("mh0", lemma)

        for lemma_key in lemma_todo:
            mh: MinHash = MinHash(num_perm = num_perm)

            for lemma in lemma_key.split(" "):
                mh.update(lemma.encode("utf-8"))

                if debug:
                    ic("mh", lemma)

            ner_mh.append(mh)

        return ner_mh


    def augment_anchor_nodes (
        self,
        chunk_nodes: dict[ str, float ],
        ner_mh: list[ MinHash ],
        *,
        num_perm: int = 128,
        debug: bool = False,
        ) -> None:
        """
Use a _locality-sensitive hash_ (LSH) to filter the entity nodes
linked to chunks, to augment the set of anchor nodes.
        """
        # index a MinHash LSH Forest of lemmatized terms
        forest: MinHashLSHForest = MinHashLSHForest(
            num_perm = num_perm,
        )

        # add lemma keys for entity nodes linked to chunks
        for iri, _ in chunk_nodes.items():
            hit: dict = self.work.ctx.erkg.get_node(iri)

            if "lemma" in hit:
                mh_hit: MinHash = MinHash(num_perm = num_perm)

                for lemma in hit["lemma"].split(" "):
                    mh_hit.update(lemma.encode("utf-8"))

                forest.add(iri, mh_hit)

        # add lemma keys for entity resolution results
        for lemma_key, ent in self.work.ctx.ent_store.entities.items():
            if ent.span.source == EntitySource.ER:
                mh_hit = MinHash(num_perm = num_perm)

                for lemma in lemma_key.split(" "):
                    mh_hit.update(lemma.encode("utf-8"))

                ent_iri: str = ent.get_iri()

                if not ent_iri in forest:
                    forest.add(ent_iri, mh_hit)

        forest.index()

        # filter the entity nodes linked to chunks, retaining those
        # closest to the `question`, to augment the anchor nodes
        lsh_top_k_question: int = self.work.config["rag"]["lsh_top_k_question"]
        lsh_top_k_lemma: int = self.work.config["rag"]["lsh_top_k_lemma"]

        for iri in forest.query(ner_mh[0], lsh_top_k_question):
            self.anchor_nodes.add(iri)

        for mh_item in ner_mh[1:]:
            for iri in forest.query(mh_item, lsh_top_k_lemma):
                self.anchor_nodes.add(iri)

        if debug:
            ic(self.anchor_nodes)


    def perform_semantic_expansion (
        self,
        *,
        w2v_top_k: int = 20,
        w2v_min_dist: float = 0.33,
        debug: bool = True,
        ) -> None:
        """
Perform a _semantic expansion_ using entity embeddings, with the set of
anchor nodes as the starting points.
        """
        decoder: dict[ int, Entity ] = self.work.ctx.ent_store.get_decoder()
        neighbors: dict[ str, float ] = {}

        for iri in self.anchor_nodes:
            try:
                anchor: dict = self.work.ctx.erkg.get_node(iri)

                if "lemma" in anchor:
                    lemma_key: str = anchor["lemma"]
                    ent: Entity = self.work.ctx.ent_store.entities[lemma_key]

                    if debug:
                        ic(iri, lemma_key)

                    for uid, distance in self.work.ctx.ent_store.w2v_model.wv.most_similar(  # type: ignore
                        str(ent.uid),
                        topn = w2v_top_k,
                    ):
                        if distance <= w2v_min_dist:
                            neigh_iri: str = decoder[int(uid)].get_iri()

                            if debug:
                                ic(neigh_iri, distance)

                            neighbors[neigh_iri] = round(distance, 4)

            except KeyError as ex:
                # TODO: using logging to trace these embedding errors?
                print("w2v_model.wv.most_similar:", iri)
                ic(ex)

        for neigh_iri, distance in sorted(neighbors.items(), key = lambda x: x[1]):
            neigh_hit: dict = self.work.ctx.erkg.get_node(neigh_iri)

            if "method" in neigh_hit and EntitySource(neigh_hit["method"]) <= EntitySource.NER:
                self.anchor_nodes.add(neigh_iri)

                if debug:
                    ic("ADD", neigh_iri, neigh_hit)


    def extract_question_subgraph (
        self,
        *,
        debug: bool = False, # True
        ) -> typing.Iterator[ str ]:
        """
Extract a subgraph, then run a _centrality_ algorithm to rerank the
most-referenced entities in the subgraph.
        """
        # TODO: visualize the walks -- the pathing does not seem to work?
        walks: set[ str ] = set(list(self.semantic_random_walk()))

        if debug:
            ic(walks)

        self.subgraph = self.work.ctx.erkg.subgraph(
            self.anchor_nodes.union(walks)
        )

        rank_iter: dict[ str, float ] = nx.pagerank(  # type: ignore
            self.subgraph,
            self.work.config["tr"]["tr_alpha"],
        ).items()

        for iri, rank in sorted(rank_iter, key = lambda x: x[1], reverse = True):  # type: ignore
            if debug:
                hit: dict = self.work.ctx.erkg.get_node(iri)  # type: ignore
                ic("tr", iri, rank, hit)  # type: ignore

            yield iri  # type: ignore


    def semantic_random_walk (
        self,
        *,
        debug: bool = True, # False
        ) -> typing.Iterator[ str ]:
        """
Generate pairwise shortest paths among the nodes from semantic
expansion, to define a subgraph.

In other words, this emulates a _semantic random walk_.
        """
        for pair in itertools.combinations(self.anchor_nodes, 2):
            if debug:
                ic(pair)

            try:
                for path in self.work.ctx.erkg.shortest_paths(pair[0], pair[1]):
                    if debug:
                        ic(path)

                    for iri in path:
                        if iri not in pair:
                            if debug:
                                hit: dict = self.work.ctx.erkg.get_node(iri)
                                ic("walk", iri, hit)

                            yield iri
            except nx.NetworkXNoPath:
                # ignore attempts when the source node is unreachable
                pass


    def find_chunk_neighbors (
        self,
        *,
        debug: bool = False,
        ) -> typing.Iterator[ int ]:
        """
Find the neighboring chunks for each anchor node.
        """
        for iri in self.anchor_nodes:
            if debug:
                ic(iri)

            for neigh_iri in self.work.ctx.erkg.neighbors(iri):
                neigh_hit: dict = self.work.ctx.erkg.get_node(neigh_iri)

                if neigh_hit.get("kind") == NodeKind.CHUNK.value:
                    chunk_id: int = TextChunk.get_uid(neigh_iri)

                    if debug:
                        ic(neigh_iri, neigh_hit, chunk_id)

                    yield chunk_id


    def get_chunks_text (
        self,
        *,
        debug: bool = True,
        ) -> list[ str ]:
        """
Retrieve text for the combined list of chunks.
        """
        # TODO: should this be sorted, i.e., as a _reranking_ function?
        chunk_ids: str = ", ".join([ str(c_id) for c_id in self.rag_chunks.keys() ])  # pylint: disable=C0201

        if debug:
            ic(chunk_ids)

        chunks: list[ str ] = self.work.ctx.chunk_table.search().where(  # type: ignore
            f"uid IN ({chunk_ids})"
        ).select(
            [ "text" ]
        ).to_polars()["text"].to_list()

        return chunks



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Manage the lexical graph.

see copyright/license https://github.com/DerwenAI/strwythura/README.md
"""

import json
import pathlib

from icecream import ic
import networkx as nx
import numpy as np
import pandas as pd

from .elem import Entity, STRW_PREFIX
from .opt import calc_quantile_bins, stripe_column, root_mean_square


class LexicalGraph:
    """
Construct a _lexical graph_ from the parsed text using  a _textgraph_
algorithm and its distillation approach called _textrank_.
This graph gets used for ranking entities, then discarded later.
    """

    def __init__ (
        self,
        config: dict,
        ) -> None:
        """
Constructor.
        """
        self.config: dict = config
        self.lex_graph: nx.MultiDiGraph = nx.MultiDiGraph()


    def increment_edge (
        self,
        edge: tuple,
        ) -> None:
        """
Increment the incidence count for a relation, creating the edge first
if needed.
        """
        if not self.lex_graph.has_edge(*edge):
            # add a relation into the graph
            self.lex_graph.add_edge(
                edge[0],
                edge[1],
                key = edge[2],
                count = 1,
            )
        else:
            # relation already exists, increment count
            self.lex_graph[edge[0]][edge[1]][edge[2]]["count"] += 1


    def add_sent (
        self,
        ent_seq: list[ Entity ],
        *,
        debug: bool = False,
        ) -> None:
        """
Augment the _textgraph_ nodes and edges from the entities co-located
in a sentence.
        """
        sem_rel: str = f"{STRW_PREFIX}follows_lexically"

        if debug:
            ic(ent_seq)

        for hop in range(self.config["tr"]["tr_lookback"]):
            for ent_pos, ent in enumerate(ent_seq[: -1 - hop]):
                rel: Entity = ent_seq[hop + ent_pos + 1]
                edge: tuple = ( ent.uid, rel.uid, sem_rel, )

                if debug:
                    ic(edge)

                if not self.lex_graph.has_node(ent.uid):
                    self.lex_graph.add_node(
                        ent.uid,
                        lemma_key = ent.lemma_key,
                        count = ent.count,
                    )

                if not self.lex_graph.has_node(rel.uid):
                    self.lex_graph.add_node(
                        rel.uid,
                        lemma_key = rel.lemma_key,
                        count = rel.count,
                    )

                self.increment_edge(edge)


    def run_textrank (
        self,
        ) -> pd.DataFrame:
        """
Run eigenvector centrality (i.e., _Personalized PageRank_) to rank the
entities using an adapted `TextRank` algorithm implementation based 
on `NetworkX` and `Polars`.
        """
        # build a dataframe of node ranks and counts
        df_rank: pd.DataFrame = pd.DataFrame.from_dict([  # type: ignore
            {
                "node_id": node_id,
                "weight": rank,
                "count": self.lex_graph.nodes[node_id]["count"],
            }
            for node_id, rank in nx.pagerank(
                    self.lex_graph,
                    alpha = self.config["tr"]["tr_alpha"],
                    weight = "count",
            ).items()
        ])

        # normalize by column and calculate quantiles
        df1: pd.DataFrame = df_rank[[ "count", "weight" ]].apply(lambda x: x / x.max(), axis = 0)
        bins: np.ndarray = calc_quantile_bins(len(df1.index))

        # stripe each columns
        df2: pd.DataFrame = pd.DataFrame([
            stripe_column(values, bins)  # type: ignore
            for _, values in df1.items()
        ]).T

        # renormalize the ranks
        df_rank["rank"] = df2.apply(root_mean_square, axis = 1)
        rank_col: np.ndarray = df_rank["rank"].to_numpy()
        rank_col /= sum(rank_col)
        df_rank["rank"] = rank_col

        return df_rank


    def load_graph (
        self,
        lex_path: pathlib.Path,
        ) -> None:
        """
De-serialize the lexical graph from a JSON file represented in the
_node-link_ data format.
        """
        with lex_path.open("r", encoding = "utf-8") as fp:
            self.lex_graph = nx.node_link_graph(
                json.load(fp),
                edges = "edges",
            )


    def save_graph (
        self,
        lex_path: pathlib.Path,
        ) -> None:
        """
Serialize the lexical graph as a JSON file represented in the
_node-link_ data format.

Aternatively this could be stored in a graph database.
        """
        with lex_path.open("w", encoding = "utf-8") as fp:
            fp.write(
                json.dumps(
                    nx.node_link_data(
                        self.lex_graph,
                        edges = "edges",
                    ),
                    indent = 2,
                    sort_keys = True,
                )
            )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Elements used for knowledge graph construction.

see copyright/license https://github.com/DerwenAI/strwythura/README.md
"""

from enum import StrEnum
import typing

from pydantic import BaseModel


STRW_BASE: str = "https://github.com/DerwenAI/strwythura/wiki/vocab#"
STRW_PREFIX: str = "strw:"


class NodeKind (StrEnum):
    """
Values for the `kind` property in graph nodes.
    """
    CHUNK = "Chunk"
    ENTITY = "Entity"
    DATAREC = "DataRec"
    TAXONOMY = "Taxonomy"


class EntitySource (StrEnum):
    """
Values for the `source` property in `Entity` instances.
    """
    TAXO = "Domain_Taxonomy"
    ER = "Entity_Resolution"
    NER = "Named_Entity_Recognition"
    NC = "Noun_Chunk"
    LEX = "Parsed_Noun"


class EntityInstance (BaseModel):
    """
Represents coordinates for one instance of an entity among the chunks.
    """
    sent_id: int
    chunk_id: int


class NounSpan (BaseModel):
    """
Represents one token span within a parsed sentence.
    """
    loc: tuple[ int, int ]
    text: str
    span: list
    label: str | None = None
    source: EntitySource = EntitySource.LEX
    iri: str | None = None


class Entity (BaseModel):
    """
Represents one entity instance:

  * defined by the domain taxonomy
  * determined by entity resolution from structured data sources
  * extracted from unstructured data sources

A non-null `iri` field indicates this entity is linked within
the constructed knowledge graph.
    """
    span: NounSpan
    lemma_key: str
    uid: int | None = None
    inst: list[ EntityInstance ] = []
    count: int = 0
    rank: float = 0.0

    def get_iri (
        self,
        ) -> str:
        """
Construct an IRI based on the `lemma_key` value.
        """
        if self.span.source == EntitySource.ER:
            return self.span.iri  # type: ignore

        stub: str = self.lemma_key.replace(" ", "_")
        return f"{STRW_PREFIX}lemma_{stub}"


def de_token_span (
    span: typing.Any,
    ) -> list[ str ]:
    """
Convert from a `spaCy` token span to a list of strings.
    """
    return list(map(str, span))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Manage the workflow components and dependency injection.

see copyright/license https://github.com/DerwenAI/strwythura/README.md
"""

import logging
import os
import pathlib
import tomllib
import typing

from icecream import ic
from sz_semantics import Thesaurus  # type: ignore
import networkx as nx

from .ctx import DomainContext, TextChunk
from .elem import Entity
from .nlp import Parser
from .scrape import Scraper


class Workflow:
    """
Manage multiple workflow phases.
    """

    def __init__ (
        self,
        *,
        config_path: pathlib.Path = pathlib.Path("config.toml"),
        ) -> None:
        """
Constructor.
        """
        # configuration
        logger: logging.Logger = logging.getLogger(__name__)  # pylint: disable=W0612
        logging.basicConfig(level = logging.WARNING) # DEBUG

        # disable noisy logging
        logging.disable(logging.ERROR)
        os.environ["TOKENIZERS_PARALLELISM"] = "0"

        with config_path.open("rb") as fp:
            self.config: dict = tomllib.load(fp)

        # dependency injection:
        #  `DomainContext` -- subclass for different domain contexts
        #  `Parser` -- subclass for different languages and dialects

        self.thesaurus: Thesaurus = Thesaurus()
        self.scraper: Scraper = Scraper(self.config)

        self.ctx: DomainContext = self.load_class(
            self.config["ctx"]["domain_class"],
            self.config,
            self.thesaurus,
        )

        self.parser: Parser | None = None


    @classmethod
    def load_class (
        cls,
        class_name: str,
        *args,
        ) -> typing.Any:
        """
Load a subclass, as a means of "plug-in" architecture for handling
different domain contexts, different languages and dialects, etc.
        """
        return globals()[class_name](*args)


    def load_parser (
        self,
        ) -> None:
        """
Load the parser, which includes the run-time cost of loading large-ish
models used in the `spaCy` NLP pipeline.
        """
        self.parser = self.load_class(
            self.config["nlp"]["parser_class"],
            self.config,
            self.ctx,
        )


    ######################################################################
    ## Part 2

    def populate_semantic_layer (
        self,
        export_path: pathlib.Path,
        domain_path: pathlib.Path,
        *,
        language: str = "en",
        ) -> None:
        """
Populate the semantic layer from the results of entity resolution.
        """
        self.thesaurus.load_source(self.thesaurus.DOMAIN_TTL)
        self.thesaurus.load_source(domain_path)

        # parse the JSON exported from Senzing ER to populate
        # a domain-specific thesaurus
        with export_path.open("r", encoding = "utf-8") as fp_json:
            for line in fp_json:
                for rdf_frag in self.thesaurus.parse_iter(line, language = language):
                    self.thesaurus.rdf_graph.parse(
                        data = self.thesaurus.RDF_PREAMBLE + rdf_frag,
                        format = "turtle",
                    )


    def build_graph_backbone (
        self,
        ) -> None:
        """
Reform the `RDFlib` semantic graph => `NetworkX` property graph
as a "backbone" for structuring the knowledge graph, prior to
entity linking from unstructured sources.
        """
        self.ctx.promote_taxo_nodes()
        self.ctx.promote_er_nodes(self.parser)
        self.ctx.promote_data_nodes()
        self.ctx.promote_er_edges()


    ######################################################################
    ## Part 3

    def make_chunks (
        self,
        chunks: list[ str ],
        chunk_size: int,
        ) -> typing.Iterator[ list[ str ] ]:
        """
Iterate through the paragraphs parsed from an article, assembling its
text chunks.
        """
        sum_chars: int = 0
        bucket: list[ str ] = []

        for text in chunks:
            num_chars: int = len(text)

            if num_chars > 0:
                if (sum_chars + num_chars) < chunk_size:
                    bucket.append(text)
                    sum_chars += num_chars
                else:
                    # emit prev bucket
                    yield bucket
                    bucket = [ text ]
                    sum_chars = num_chars

        # emit last bucket
        yield bucket


    def crawl_chunk_parse (
        self,
        content_sources: list[ str ],
        *,
        debug: bool = False,
        ) -> None:
        """
For each of the given URLs:

  - load the document's HTML, through a cache
  - extract the document text as a sequence of paragraphs
  - create a `TextChunk` for each paragraph
  - add the chunk and its embedding to the vector store
  - parse the text in each chunk, linking into the graph
        """
        chunk_size: int = self.config["nlp"]["chunk_size"]

        for url in content_sources:
            sent_id: int = 0
            ic(url)

            for chunks in self.make_chunks(self.scraper.scrape_html(url), chunk_size):
                # add each accumulated text chunk and its embedding
                # to the vector store
                chunk: TextChunk = self.ctx.add_chunk(
                    url,
                    sent_id,
                    "\n\n".join(chunks),
                )

                if debug:
                    ic(chunk)

                # parse each paragraph
                for para in chunks:
                    num_sent: int = self.parser.parse_para(  # type: ignore
                        chunk.uid,
                        para,
                        debug = debug,
                    )

                    if debug:
                        ic(para, num_sent)

                    sent_id += num_sent


    ######################################################################
    ## Part 5

    def distill_knowledge_graph (
        self,
        ) -> None:
        """
Finalize construction and serialization of the knowledge graph after
entity linking:

  - run _textrank_ on the lexical graph to rank entities by weight and count
  - calculate conditional probability of entity _co-occurrence_ within a sentence
  - promote distilled entities into the knowledge graph
  - cross-link entities and chunks within the knowledge graph
  - serialize both graphs for later use in inference
        """
        # run the `textrank` algorithm, approximating a hypervolume of
        # the `weight` and `count` to normalize as a ranking, then
        # update the rank values in both the lexical graph and the
        # entity store
        for _, row in self.ctx.lex.run_textrank().iterrows():
            node_id: int = row["node_id"]
            rank: float = round(float(row["rank"]), 5)

            nx.set_node_attributes(
                self.ctx.lex.lex_graph,
                { node_id : rank },
                "rank",
            )

            lemma_key: str = self.ctx.lex.lex_graph.nodes[node_id]["lemma_key"]
            ent: Entity = self.ctx.ent_store.entities[lemma_key]
            ent.rank = rank

        # cross-link entities and chunks
        self.ctx.link_entity_chunks()

        # calculate the conditional probability for each pair of
        # entities which co-occur in a sentence, then represent
        # in the `LexicalGraph` using semantic relations
        self.ctx.co_occur_entities()

        # promote distilled entities into the knowledge graph
        self.ctx.promote_ner_nodes()


    ######################################################################
    ## Part 7

    def load_assets (
        self,
        ) -> None:
        """
De-serialize assets from the previous steps.
        """
        self.thesaurus.load_source(
            pathlib.Path(self.config["sz"]["thesaurus_path"]),
            format = "turtle",
        )

        self.ctx.ent_store.load_json(
            pathlib.Path(self.config["ent"]["store_path"]),
        )

        self.ctx.ent_store.load_w2v(
            pathlib.Path(self.config["ent"]["w2v_path"]),
        )

        self.ctx.lex.load_graph(
            pathlib.Path(self.config["nlp"]["lex_path"]),
        )

        self.ctx.erkg.load_graph(
            pathlib.Path(self.config["erkg"]["erkg_path"]),
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Manage the entity store.

see copyright/license https://github.com/DerwenAI/strwythura/README.md
"""

from collections import OrderedDict
import json
import pathlib

from arrowspace import ArrowSpaceBuilder, GraphLaplacian  # type: ignore  # pylint: disable=E0611
from icecream import ic
import gensim  # type: ignore
import numpy as np

from .elem import Entity, \
    de_token_span


class EntityStore:
    """
Manage the semantic graph embeddings with this domain context:
    vocabulary => taxonomy => thesaurus => ontology
    """

    def __init__ (
        self,
        config: dict,
        ) -> None:
        """
Constructor.
        """
        self.config: dict = config

        # manage the known entities
        self.entities: OrderedDict = OrderedDict()

        # entity UIDs used for embedding, etc.
        self.max_uid: int = 0

        # entity embeddings
        self.w2v_vectors: list[ list[ int ] ] = []
        self.w2v_model: gensim.models.Word2Vec | None = None


    def increment_uid (
        self,
        ) -> int:
        """
Increment the count of nodes.
        """
        uid: int = self.max_uid
        self.max_uid += 1

        return uid


    def get_decoder (
        self,
        ) -> dict[ int, Entity ]:
        """
Construct a dictionary to lookup entities by their `uid` value.
        """
        return {
            ent.uid: ent
            for ent in self.entities.values()
        }


    def encode_entity (
        self,
        ent: Entity,
        *,
        create: bool = False,
        ) -> Entity | None:
        """
Encode a known entity, indexed by its parsed lemma key in the entity
store, and set its IRI as a unique identifier in the ERKG.

This encoding serves as a UID for nodes in the semantic layer and
within vector representation for embeddings.

If a entity arrives with a duplicate `lemma_key` then promote the
one which comes from a higher priority source (and therefore more
info populated).

Return the entity which is stored.
        """
        if create:
            if ent.lemma_key not in self.entities:
                # add a new entity into the store
                ent.uid = self.increment_uid()
                ent.count = 1
                self.entities[ent.lemma_key] = ent

            elif ent.span.source < self.entities[ent.lemma_key].span.source:
                # replace previous entity with one from a a higher priority source
                ent.uid = self.entities[ent.lemma_key].uid
                ent.count = self.entities[ent.lemma_key].count + 1
                self.entities[ent.lemma_key] = ent

            else:
                # increment the incidence count
                self.entities[ent.lemma_key].count += 1

        return self.entities.get(ent.lemma_key)


    def load_json (
        self,
        store_path: pathlib.Path,
        ) -> None:
        """
De-serialize the entity definitions from a JSONL file, first clearing
any previous definitions.
        """
        self.entities = OrderedDict()

        with store_path.open("r", encoding = "utf-8") as fp:
            for line in fp:
                ent: Entity = Entity.model_validate(
                    json.loads(line),
                )

                self.entities[ent.lemma_key] = ent
                self.max_uid = max(self.max_uid, ent.uid + 1)  # type: ignore

        ic(self.max_uid, len(self.entities))


    def save_json (
        self,
        store_path: pathlib.Path,
        ) -> None:
        """
Serialize the entity definitions to a JSONL file.
        """
        with store_path.open("w",  encoding = "utf-8") as fp:
            for ent in self.entities.values():
                ent.span.span = de_token_span(ent.span.span)
                fp.write(json.dumps(ent.model_dump(mode = "json")))
                fp.write("\n")


    def embed_sequence (
        self,
        seq_vec: list[ int ],
        ) -> None:
        """
Build an embedding vector input for a sequence of entities from a
parsed sentence.
        """
        self.w2v_vectors.append(seq_vec)


    def load_vec (
        self,
        vec_path: pathlib.Path,
        ) -> None:
        """
De-serialize the entity embedding vectors from a text file,
overwriting any previous definitions.
        """
        with vec_path.open("r", encoding = "utf-8") as fp:
            self.w2v_vectors = []

            for line in fp:
                self.w2v_vectors.append(
                    [ int(x) for x in line.strip().split(",") ]
                )


    def save_vec (
        self,
        vec_path: pathlib.Path,
        ) -> None:
        """
Serialize the entity embedding vectors to a text file.
        """
        with vec_path.open("w", encoding = "utf-8") as fp:
            for vec in self.w2v_vectors:
                vec_rep: str = ",".join(map(str, vec))

                # filter null vectors
                if len(vec_rep) > 0:
                    fp.write(vec_rep)
                    fp.write("\n")


    def train_embeddings (
        self,
        ) -> None:
        """
Train a `gensim.Word2Vec` model for entity embeddings.
        """
        w2v_max: int = max([  # pylint: disable=R1728
            len(vec) - 1
            for vec in self.w2v_vectors
        ])

        # TODO: here's where we need to resolve once synonyms get
        # introduced into the thesaurus via curation
        w2v_vect: list[list[ str ]] = [
            [
                str(x)
                for x in vec
            ]
            for vec in self.w2v_vectors
        ]

        self.w2v_model = gensim.models.Word2Vec(
            sentences = w2v_vect,
            vector_size = 23,
            window = w2v_max,
            min_count = 1,
            sg = 1, # use skip-gram, not CBOW
        )


    def load_w2v (
        self,
        w2v_path: pathlib.Path,
        ) -> None:
        """
De-serialize the entity embeddings model from a file in `gensim.Word2Vec`
format.
        """
        self.w2v_model = gensim.models.Word2Vec.load(w2v_path.as_posix())


    def save_w2v (
        self,
        w2v_path: pathlib.Path,
        ) -> None:
        """
Serialize the entity embeddings model to a file in `gensim.Word2Vec`
format.
        """
        self.w2v_model.save(w2v_path.as_posix())  # type: ignore


    def build_aspace (
        self,
        w2v_model: gensim.models.Word2Vec,
        *,
        tau: float = 1.0,
        debug: bool = True,
        ) -> tuple[ ArrowSpaceBuilder, GraphLaplacian ]:
        """
Build an `ArrowSpace` computed signal graph and lambdas,
then compare with `gensim` similarity measures.
        """
        # extract the embedding vectors from a `gensim` model
        embed_vecs: list = []

        for uid, lemma_key in enumerate(self.entities.keys()):
            entity_key: str = str(uid)

            if entity_key in w2v_model.wv.index_to_key:
                embedding: list[ float ] = w2v_model.wv[entity_key]
                embed_vecs.append(embedding)

                if debug:
                    ic(uid, lemma_key)
                    print(embedding)

        # build an ArrowSpace with computed signal graph and lambdas
        aspace_params: dict = {
            "eps": 1.0,
            "k": 6,
            "topk": 6,
            "p": 2.0,
            "sigma": 1.0,
        }

        aspace, gl = ArrowSpaceBuilder.build(
            aspace_params,
            np.array(
                embed_vecs,
                dtype = np.float64,
            ),
        )

        # search comparable items
        # defaults: k = nitems, alpha = 0.9, beta = 0.1
        for query_uid in [ ent.uid for ent in self.entities.values() ]:
            query: np.ndarray = np.array(
                embed_vecs[query_uid],
                dtype = np.float64,
            )

            similar_words = w2v_model.wv.most_similar(
                str(query_uid),
                topn = 5,
            )

            if debug:
                ic(similar_words)

                for uid, sim_metric in aspace.search(query, gl, tau):
                    if uid != query_uid:
                        ic(uid, sim_metric)

        # return the ArrowSpace and Graph laplacian
        return aspace, gl
